/***************
 *  UGPCONV2.C                    by David Khling
 ***************
 *
 * This file uses UGP.C to convert .TGA files to UGP or raw pictures, to be
 * displayed on the TI-92 calculator with the assembly shell Fargo (by
 * David Ellsworth).
 *
 * To display UGP pictures, use the library UGPLIB.92p (that was programmed
 * by me)
 *
 * Do with this file whatever you want. But I would appreciate if you'd
 * put extensions etc. into an extra module that can be linked with ugpconv.c.
 * If you'd send me this module I will add it to MY version of ugpconv.c. By
 * this I will be able to add the new features to future releases.
 *
 * This file can only be compiled with a DJGPP C compiler, due to the low level
 * graphics functions. This is a free software compiler you can download from
 * http://www.delorie.com. If you don't have this compiler, comment out all
 * display routines, and the include "vga256.c" command.
 *
 * You can contact me via email: dvdkhlng <at> gmx <dot> de.
 *
 * Last modification: december 1998
 *
 * This file is in the public domain.
 */

/** number of current version */
#define VERSION_STR "2.10"

/***************************************************************************/
#include <stdio.h>
#include <conio.h>
#include <string.h>
#include <errno.h>
#include <dir.h>
#include <signal.h>
#include "ugp.h"

/* Primitive, but small graphic functions of my brother (Felix) */
#include "vga256.c"
/***************************************************************************/

/***************************************************************************/
BASE_PIC     *Picture[255];     /* unconverted pictures (8bit/pixel) */
LOWCOLOR_PIC *ConvPic[255];     /* lowcolor - converted pictures */

char     OutPicture[256] = "";  /* name of output file */

int      PictureNum,            /* number of specified pictures */
         Compressed   = 0,      /* compress ugp picture/animation */
         Interleaved  = 0,      /* store picture(s) interleaved */
         SlideShow    = 0,      /* create slide show instead of animation */
         Raw          = 0,      /* store picture(s) raw */
         GraphLink92s = 0,      /* output ugp pictures into a 92s file */
         XOR_Ani      = 0,      /* store animation XORed */
         Planes       = 2,      /* number of bitplanes of dest. pic */
         Dithering    = 1,      /* dithering (1 = ordered dither 1) */
         Brightness   = 128,    /* brightness (for convertion) */
         Contrast     = 128,    /* contrast (for convertion */
         DispPics     = 0,      /* display pictures */
         OldCursorX,            /* cursor pos., before graphic was displayed */
         OldCursorY;

void     *TextScreen;           /* buffer to save the text screen */
/***************************************************************************/


/**
 * Initializes the graphic display. Stores the contents of the text screen
 * to >TextScreen<.
 */
void InitDisplay (int planes)
{
  int I;
  LCD_COLOR *color;

  /** save screen contents and cursor pos */
  ScreenGetCursor (&OldCursorY, &OldCursorX);
  ScreenRetrieve (TextScreen = xmalloc (ScreenRows()*ScreenCols()*2));

  /** set screenmode 13h */
  Init13 ();

  /** set the background color */
  SetOneDAC13 (0, 8, 8, 8);
  Clear13 (0); /* set screen to background color */

  /** set the display default color (white) */
  SetOneDAC13 (16, TI92_LCD_White.r>>2, TI92_LCD_White.g>>2, TI92_LCD_White.b>>2);

  /** set the palette, to display the TI-92 LCD colors correctly */
  for (I = 0; I < ColorNum[planes]; I++)
  {
    color = getColorRGB (I, planes);
    /** set the color - color is in range 0..255, but 0..63 is expected */
    SetOneDAC13 (I+1, color->r>>2, color->g>>2, color->b>>2);
    Free_LCD_COLOR (color);
  }

  /** paint the TI-92 display */
  Box13 (40, 36, 240, 128, 16);
}

/**
 * Closes the graphic display, and restores the contents of the text screen.
 */
void CloseDisplay ()
{
  Close13 ();
  ScreenUpdate (TextScreen);
  free (TextScreen);
  ScreenSetCursor (OldCursorY, OldCursorX);
}

/**
 * Displays a color-depth reduced picture, centered
 */
void Display_LOWCOLOR_PIC (LOWCOLOR_PIC *pic)
{
  int       Ix, Iy,
            Xpos, Ypos;
  char      *pixel = pic->Pic;

  /** left, top corner of picture, to display it centered */
  Xpos = 160 - pic->Width/2;
  Ypos = 100 - pic->Height/2;

  for (Iy = 0; Iy < pic->Height; Iy++)
    for (Ix = 0; Ix < pic->Width; Ix++)
      Put13 (Ix + Xpos, Iy + Ypos, *pixel++ + 1);
}

void Abort (int dummy)
{
  puts ("\nI'm immortal! :-)");
  return;
}

/**
 * Displays a command line option summary
 */
void Help ()
{
  char *helptext[] =
  {
    "UGPCONV [file1] [file2] [file3] [...] [option1] [option2] [...]",
    "  [fileX]: Input 24bit uncompressed TGA file(s) (can contain wildcards)",
    "  [optionX]:",
    "  /o<filename>   specify an output file",
    "  /p<num>        set the number of bitplanes 1..3 (default: 2)",
    "  /r             output raw file(s) instead of one UGP file",
    "  /s             output ugp file + title + ani delay into a 92s file",
    "  /S             show a preview of the picture(s)",
    "  /D<num>        set the dither mode to <num> (default: 1)",
    "                   0: no dithering (nearest solid)",
    "                   1: ordered dither 1 (small pattern)",
    "                   2: ordered dither 2 (big pattern)",
    "                   3: ordered dither 3 (huge pattern)",
    "                   4: error diffusion  (random pattern)",
    "  /C<num>        change the contrast   0..255 (default: 128)",
    "  /B<num>        change the brightness 0..255 (default: 128)",
    "  /c             (not RAW) compress picture(s)",
    "  /i             (not RAW) store picture(s) interleaved",
    "  /x             (not RAW) store animation pictures XOR to their predecessor",
    "  /l             (not RAW) create slide show instead of animation",
    "",
    "In UGP mode, if one picture is specified the output will be a UGP picture,",
    "else an UGP animation is created.",
    NULL
  };
  int I;

  for (I = 0; helptext[I]; I++)
    puts (helptext[I]);

  exit (0);
}

/**
 * Prints the "Invalid argument:..." error and aborts the prog.
 */
void InvalidArg (char *arg)
{
  fprintf (stderr, "\n\nInvalid argument: \"%s\"\n\n", arg);
  Help ();
  exit (1);
}


/****
 * MAIN - processes command line options, and converts the pics.
 */
int main (int argn, char *argv[])
{
  int  I;

  /** install Ctrl+C / Ctrl+Pause handler */
  signal (SIGINT, Abort);

  puts ("\n\t---------------------------------------------------------------");
  puts (  "\t                        UGPCONV "VERSION_STR"");
  puts (  "\t                   by David Khling 1998                       ");
  puts (  "\t---------------------------------------------------------------\n");

  /** process options */
  for (I = 1, PictureNum = 0; I < argn; I++)
    if (argv[I][0] == '-' || argv[I][0] == '/')
      switch (argv[I][1])
      {
        case 'c': if (argv[I][2]) InvalidArg (argv[I]);
                  Compressed = 1;  break;
        case 'i': if (argv[I][2]) InvalidArg (argv[I]);
                  Interleaved = 1; break;
        case 'x': if (argv[I][2]) InvalidArg (argv[I]);
                  XOR_Ani = 1; break;
        case 'l': if (argv[I][2]) InvalidArg (argv[I]);
                  SlideShow = 1; break;
        case 'r': if (argv[I][2]) InvalidArg (argv[I]);
                  Raw = 1; break;
        case 's': if (argv[I][2]) InvalidArg (argv[I]);
                  GraphLink92s = 1; break;
        case 'D': if (!sscanf (&argv[I][2], "%i", &Dithering) ||
                      Dithering < NEAREST_SOLID || Dithering > ERROR_DIFF)
                    InvalidArg (argv[I]);
                  break;
        case 'B': if (!sscanf (&argv[I][2], "%i", &Brightness) ||
                      Brightness < 0 || Brightness > 255)
                    InvalidArg (argv[I]);
                  break;
        case 'C': if (!sscanf (&argv[I][2], "%i", &Contrast) ||
                      Contrast < 0 || Contrast > 255)
                    InvalidArg (argv[I]);
                  break;
        case 'p': if (!sscanf(&argv[I][2], "%i", &Planes) ||
                      Planes < 0 || Planes > 3)
                    InvalidArg (argv[I]);
                  break;
        case 'o': strcpy (OutPicture, &argv[I][2]);
                  break;
        case 'S': if (argv[I][2]) InvalidArg (argv[I]);
                  DispPics = 1; break;
        case '?':
        case 'h': Help ();
                  break;
        default:  InvalidArg (argv[I]);
      }
    else
      PictureNum++;

  /** check options */
  if (!OutPicture[0])
  {
    fputs ("\n\nNo output file specified!\n", stderr);
    return (1);
  }
  if (XOR_Ani && PictureNum <= 1)
  {
    fputs ("\n\nCan't create XOR animation of one picture\n", stderr);
    return (1);
  }
  if (SlideShow && PictureNum <= 1)
  {
    fputs ("\n\nCan't create slide show of one picture\n", stderr);
    return (1);
  }
  if (Raw && (Compressed || XOR_Ani || SlideShow))
  {
    fprintf (stderr, "\n\nRaw pictures can neither be XOR, nor compressed, nor slide show.\n");
    return (1);
  }

  /** read files */
  for (I = 1, PictureNum = 0; I < argn; I++)
    if (argv[I][0] != '-' && argv[I][0] != '/')
    {
      printf ("Loading \"%s\"...\n", argv[I]);
      if ((Picture[PictureNum] = Load_TGA_24bit_uncompressed (argv[I])) == NULL)
      {
        fprintf (stderr, "\nLoad error: %s\n\n", LoadErrorStr[LoadError]);
        return (1);
      }
      if (Picture[PictureNum]->Width != Picture[0]->Width ||
          Picture[PictureNum]->Height != Picture[0]->Height)
      {
        fputs ("\nLoad error: pictures differ in their size\n\n", stderr);
        return (1);
      }
      PictureNum++;
    }

  /** convert picture(s) */
  for (I = 0; I < PictureNum; I++)
  {
    printf ("\rDecreasing color depth... %3i%%", I*100 / PictureNum);
    fflush (stdout);
    ConvPic[I] = ColorConv (Picture[I], Brightness, Contrast, Planes, Dithering);

    /* remove base picture */
    Free_BASE_PIC (Picture[I]);
  }
  puts ("\rDecreasing color depth... 100%");

  /** display the pictures */
  if (DispPics)
  {
    InitDisplay (Planes);
    do {
      for (I = 0; I < PictureNum; I++)
      {
        Display_LOWCOLOR_PIC (ConvPic[I]);
        if (SlideShow)
          getch ();
        else
          delay (100);
      }
    } while (!kbhit () && !SlideShow);
    if (!SlideShow)
      while (kbhit ()) getch ();  /* empty keyboard buffer */

    CloseDisplay ();
  }

  /** store picture(s) */
  if (Raw)   /* store picture(s) raw */
  {          /* if picnum > 1: store all raw pictures into numbered files */

    char drive[MAXDRIVE], path[MAXDIR], file[MAXFILE], ext[MAXEXT];
    long picsize;

    /* split OutPicture, to generate numbered filenames */
    fnsplit(OutPicture, drive, path, file, ext);

    if (PictureNum > 1)
      file[5] = '\0'; /* cut filename down to 5 chars */

    for (I = 0; I < PictureNum; I++)
    {
      RAW_PIC *raw;
      char    newfile[MAXFILE], /* name of dest. file (without path, ext. etc) */
              newname[MAXPATH]; /* complete filename of dest. file */

      if (PictureNum > 1)     /* concatenate number to filename */
        sprintf (newfile, "%s%03i", file, I);
      else
        strcpy (newfile, file);
      fnmerge (newname, drive, path, newfile, ext);

      printf ("Storing raw picture #%03i to \"%s\"...\n", I, newname);
      raw = Create_RAW_PIC (ConvPic[I], Interleaved);
      picsize = RAW_SIZE (raw);

      if (Store_RAW_PIC (raw, newname))
      {
        fprintf (stderr, "\n\nCan't create \"%s\": %s\n", newname, strerror (errno));
        return (1);
      }
    }
    printf ("Size of one raw picture: %i bytes; together: %li bytes\n",
            picsize, (long)picsize * PictureNum);
  }
  else       /* store picture(s) into one ugp - file */
  {
    UGP_PIC *ugp;

    printf ("Converting picture%s to UGP...\n", PictureNum > 1 ? "s" : "");
    if ((ugp = Create_UGP_PIC (ConvPic, PictureNum, Compressed, Interleaved, XOR_Ani, SlideShow)) == NULL)
    {
      fputs ("\n\nError during the UGP convertion.\n", stderr);
      return (1);
    }

    if (GraphLink92s) /* store picture into a 92s - string var */
    {
      char title[256] = "";
      int  anidelay = 0;

      printf ("Title of UGP-STR picture: ");
      gets (title);
      title[30] = '\0';         /* cut title down to 30 chars */
      if (PictureNum > 1 && !SlideShow)
      {
        printf ("Delay time per picture (1 = 1/350 secs): ");
        scanf ("%i", &anidelay);
      }
      printf ("Storing UGP %s to Graph Link String variable \"%s\"...\n",
               PictureNum > 1 ? "animation" : "picture", OutPicture);
  
      if (Store_UGP_PIC_to_92S (ugp, title, anidelay, OutPicture))
      {
        fprintf (stderr, "\n\nCan't create \"%s\": %s\n", OutPicture, strerror (errno));
        return (1);
      }
    }
    else
    {
      printf ("Storing UGP %s to \"%s\"...\n",
        PictureNum > 1 ? (SlideShow ? "slideshow" : "animation") : "picture", OutPicture);
  
      if (Store_UGP_PIC (ugp, OutPicture))
      {
        fprintf (stderr, "\n\nCan't create \"%s\": %s\n", OutPicture, strerror (errno));
        return (1);
      }
    }

    if (Compressed)
      printf ("File compressed to %i%% of original size.\n", CompressionOpt);
    printf ("Size of UGP file: %li bytes\n", UGP_SIZE (ugp));
    Free_UGP_PIC (ugp);
  }

  return (0);
}
